#version 120
//this code was derived by eiffie from Inigo Quilez's article on brute force path tracing found here:
//http://iquilezles.org/www/articles/
//added to that code by eiffie are examples of refraction and distance estimated lighting including specular

#info global illum
//this was built with 3D.frag as of v0.9.12b2
#include "3D.frag" 

#group Colors
uniform vec3 sunColor;color[1.0,1.0,0.5]
uniform vec3 skyColor;color[0.3,0.6,1.0]
uniform vec3 bulbColor;color[0.75,0.6,0.9]
uniform float bulbReflect;slider[-1.0,-1.0,1.0]
uniform float bulbSpec;slider[0.0,1.5,2.0]
uniform int bulbSpecExp;slider[1,128,512]
uniform vec3 mengerColor;color[0.75,0.75,0.75]
uniform float mengerReflect;slider[-1.0,0.2,1.0]
uniform float mengerSpec;slider[0.0,0.0,2.0]
uniform int mengerSpecExp;slider[1,16,512]
#group Positions
uniform vec3 sunDirection;slider[(-1.0,-1.0,-1.0),(0.25,1.0,-0.5),(1.0,1.0,1.0)]
uniform vec3 bulbPosition;slider[(-1.0,-1.0,-1.0),(0.0,0.0,0.5),(1.0,1.0,1.0)]
uniform vec3 lightPosition;slider[(-1.0,-1.0,-1.0),(0.25,0.25,-0.25),(1.0,1.0,1.0)]
#group Emissive
uniform bool UseEmissive; checkbox[false]
uniform vec3 lightColor;color[1.0,0.8,0.75]
uniform float lightStrength;slider[0.0,0.9,1.0]
uniform float lightBloom;slider[0.0,0.5,1.0]
#group Raytracer
uniform int RayBounces;slider[0,6,10]
uniform int RaySteps;slider[64,128,512]
uniform int ShadowSteps;slider[32,96,512]
uniform float FudgeFactor;slider[0.1,0.9,1.0]
uniform float HitDistance;slider[0.0001,0.001,0.001]
uniform bool UseCheapSpecular;checkbox[false]

const float maxDepth=10.0;	//depth of scene
int iSteps=RaySteps;//taking the user input and making it ready to use
float lStrength=2000.0-pow(lightStrength,0.001)*2000.0;
float lBloom=5000000.0-pow(lightBloom,0.01)*5000000.0;
float minLiteDist,side=1.0; //1.0=outside, -1.0=inside of object
vec3 sunDir=normalize(sunDirection);

//some simple distance estimate functions
float DESphere(in vec3 z, float radius){return length(z)-radius;}
float DEBox(in vec3 z, float hlen){return max(abs(z.x),max(abs(z.y),abs(z.z)))-hlen;}
float DECylinder(in vec3 z, float hlen, float radius){return max(length(z.zy)-radius,abs(z.x)-hlen);}

const float scale=3.0;//menger constants
const vec3 offset=vec3(1.0,1.0,1.0);
const int iters=5;
const float psni=pow(scale,-float(iters));

float DEMenger(in vec3 z)
{
	for (int n = 0; n < iters; n++) {
		z = abs(z);
		if (z.x<z.y)z.xy = z.yx;
		if (z.x<z.z)z.xz = z.zx;
		if (z.y<z.z)z.yz = z.zy;
		z = z*scale - offset*(scale-1.0);
		if(z.z<-0.5*offset.z*(scale-1.0))z.z+=offset.z*(scale-1.0);
	}
	return DEBox(z,scale*0.5)*psni;
}

float DEBulb(in vec3 z0, int iters, float scale, float power)
{
	//return length(z0)-1.0/scale;
	vec3 c = z0*scale,z =c;
	float dr = scale,r = length(z),zr,zo,zi;
	for (int n = 0; n < iters && r<2.0; n++) {
		zo = asin(z.z / r) * power;
		zi = atan(z.y, z.x) * power;
		zr = pow(r, power-1.0);
		dr = dr * zr * power + 1.0;
		z=(r*zr)*vec3(cos(zo)*vec2(cos(zi),sin(zi)),sin(zo))+c;
		r = length(z);
	}
	return 0.5 * log(r) * r / dr;
}

float DELite(in vec3 z0)//This is light object. It can be any continuous field with folds etc.
{
	z0-=lightPosition;
	return DEBulb(z0,1,16.0,4.0);
	//below is an example of creating multiple lights
	//vec3 z=z0+vec3(0.0,0.0,0.75+cos(z0.x*0.5+z0.y*0.25+lightPosition*1.57+1.57)*0.25);
	//for (int n = 0; n < 2; n++) z = abs(z)*3.0 - vec3(2.0,2.0,0.0);
	//return (length(z.xyz)-0.1)*0.1;
}

//you need to implement this function!!!
vec2 map(in vec3 pos)
{//return distance estimate and object id
	float flr=max(pos.y+1.0,length(pos.xz)-5.0);
	float bulb=DESphere(pos-bulbPosition,0.25);//DEBulb(pos-bulbPosition,4,4.0,5.0);
	float mgr=DEMenger(pos);
	float lit=(UseEmissive)?DELite(pos):1000.0;
	minLiteDist=min(minLiteDist,lit);//save the closest distance to the light for bloom
	float id=1.0; //just sorting and assigning id
	if(lit<mgr)id=-2.0; //-2.0 is the id for emmisive light
	else lit=mgr;
	if(flr<lit){lit=flr;id=2.0;}
	return vec2(min(bulb,lit),(bulb<lit)?0.0:id);
}

// you don't need to change the following unless you want to :)
float mapL(in vec3 pos){return DELite(pos);}

vec3 getColor( in vec3 pos, in vec3 nor, in float item )
{//get the color of the surface at a point based on the item hit
	return (item==0.0)?bulbColor:(item==1.0)?mengerColor:vec3(0.2);
}
vec3 getMaterial(in float item)
{//get a vector (reflectivity,specularity,specularExp) for the object
	return (item==0.0)?vec3(bulbReflect,bulbSpec,bulbSpecExp):(item==1.0)?vec3(mengerReflect,mengerSpec,mengerSpecExp):vec3(0.0,0.0,1.0);
}

vec3 getBackground( in vec3 rd ){return skyColor+sunColor*(max(0.0,dot(rd,sunDir))*0.2+pow(max(0.0,dot(rd,sunDir)),256.0));}

vec2 intersect(in vec3 ro, in vec3 rd )
{//march the ray until you hit something or go out of bounds
	float t=HitDistance*5.0;
	side=sign(map(ro+t*rd).x);
	vec2 h=vec2(-1.0),mult=vec2(side*FudgeFactor,1.0);
	for(int i=0;i<iSteps && t<maxDepth;i++){
		h=map(ro+t*rd)*mult;
		t+=h.x;
		if(abs(h.x)<HitDistance)return vec2(t,h.y);
	}//returns distance, object id
	return vec2(t,(t<maxDepth)?h.y:-1.0);//-1.0 is id for "hit nothing"
}

vec3 ve=vec3(HitDistance*0.1,0.0,0.0);
vec3 getNormal( in vec3 pos, in float item )
{// get the normal to the surface at the hit point
	return side*normalize(vec3(-map(pos-ve.xyy).x+map(pos+ve.xyy).x,
		-map(pos-ve.yxy).x+map(pos+ve.yxy).x,-map(pos-ve.yyx).x+map(pos+ve.yyx).x));
}

vec3 shadow(in vec3 ro, in vec3 rd, float liteType )
{//if we hit a highly transparent or reflective object we bounce the ray
	vec3 shad=vec3(1.0);
	for(int i=0;i<RayBounces && dot(shad,shad)>0.001;i++){
		vec2 h=intersect( ro, rd );
		if(h.y<0.0)return vec3(float(h.y==liteType));
		else {//for materials with high refraction or reflection keep bouncing
			float r=getMaterial(h.y).x;
			if(r>=-0.5 && r<0.5) return vec3(0.0);
			ro+=rd*h.x;
			vec3 nor = getNormal( ro, h.y );
			shad*=getColor(ro, nor, h.y)*abs(r);
			vec3 rdNew=(side>=0.0)?refract(rd,nor,0.75):refract(rd,nor,1.33);
			if(dot(rdNew,nor)<-0.05 && r<0.0)rd=rdNew;
			else rd=reflect(rd,nor);
		}
	}
	return shad;
}

//random seed and generator
vec2 randv2=fract(cos((gl_FragCoord.xy+gl_FragCoord.yx*vec2(1000.0,1000.0))+vec2(backbufferCounter,backbufferCounter))*10000.0);
vec2 rand2(){// implementation derived from one found at: lumina.sourceforge.net/Tutorials/Noise.html
	randv2+=vec2(1.0,1.0);
	return vec2(fract(sin(dot(randv2.xy ,vec2(12.9898,78.233))) * 43758.5453),
		fract(cos(dot(randv2.xy ,vec2(4.898,7.23))) * 23421.631));
}
 
vec3 cosPowDir(vec3  dir, float power, float extent) {	
	vec2 r=rand2()*vec2(6.2831853,extent);
	vec3 sdir=cross(dir,((abs(dir.x)<0.5)?vec3(1.0,0.0,0.0):vec3(0.0,1.0,0.0)));
	vec3 tdir=cross(dir,sdir);
	r.y=pow(1.0-r.y,1.0/power);
	float oneminus = sqrt(1.0-r.y*r.y);
	return cos(r.x)*oneminus*sdir + sin(r.x)*oneminus*tdir + r.y*dir;
}

vec3 applyLighting( inout vec3 fcol, in vec3 rd, in vec3 pos, in vec3 nor, in float item)
{//calculate ambient, emissive and specular lighting
	float saveSide=side,saveMLD=minLiteDist;iSteps=ShadowSteps;
	vec3 scol=vec3(0.0),ldir=cosPowDir(nor,1.0,1.0);
	vec3 dcol=getBackground(ldir)*shadow( pos, ldir, -1.0 );//check for sky light
	ldir=cosPowDir(sunDir,1.0,0.001);
	if(dot(ldir,nor)>0.0){//check for sunlight
		vec3 ecol=shadow( pos, ldir, -1.0 );
		if(dot(ecol,ecol)>0.0){
			dcol += dot(ldir, nor) * ecol * sunColor;
			if(UseCheapSpecular)scol+=getMaterial(item).y*fcol*sunColor*pow(max(0.0,dot(rd,reflect(ldir,nor))),getMaterial(item).z);
		}
	}//below brings in specular light from every direction???
	if(!UseCheapSpecular && getMaterial(item).y>0.0){
		ldir=cosPowDir(reflect(rd,nor),getMaterial(item).z,1.0);
		if(dot(ldir,sunDir)>0.9)scol+=getMaterial(item).y*fcol*sunColor*shadow(pos, ldir, -1.0);
	}
	if(UseEmissive){
		vec2 vt=vec2(mapL(pos),0.0);//find emissive light dir by triangulating its nearest point
		ldir =-normalize(vec3(-mapL(pos-vt.xyy)+mapL(pos+vt.xyy),
			-mapL(pos-vt.yxy)+mapL(pos+vt.yxy),-mapL(pos-vt.yyx)+mapL(pos+vt.yyx)));
		vec3 lcol=lightColor/max(lStrength*vt.x*vt.x,1.0);
		if(!UseCheapSpecular && getMaterial(item).y>0.0){
			ldir=cosPowDir(reflect(rd,nor),getMaterial(item).z,1.0);
			scol+=getMaterial(item).y*fcol*lcol*shadow(pos, ldir, -2.0);
		}
		ldir=cosPowDir(ldir,1.0,0.001/max(vt.x*vt.x,1.0));
		if(dot(ldir,nor)>0.0){//check for emissive and specular light
			vec3 ecol=shadow( pos, ldir, -2.0 );
			if(dot(ecol,ecol)>0.0){//we hit the light so calc specular light
				dcol+=lcol*dot(nor,ldir)*ecol;
				if(UseCheapSpecular)scol+=getMaterial(item).y*fcol*lcol*pow(max(0.0,dot(rd,reflect(ldir,nor))),getMaterial(item).z);
			}//above is the quick hack for specular
		}
	}
	side=saveSide;minLiteDist=saveMLD;iSteps=RaySteps;//we need to return these globals to their old values (silly coder)
	fcol *= getColor( pos, nor, item );//updating the running color value
	return fcol*dcol+scol;
}

vec3 getBRDFRay( in vec3 nor, in vec3 rd, in float item )
{//randomly direct the ray in a hemisphere or cone based on reflectivity
	float r=getMaterial(item).x;
	if( rand2().x > abs(r)) return cosPowDir( nor,1.0,1.0 );
	else {//return a cone direction for a reflected or refracted ray
		vec3 p=reflect(rd,nor);
		if(r<0.0){//refract or reflect perfectly
			if(rand2().x<0.75){
				rd=(side>=0.0)?refract(rd,nor,0.75):refract(rd,nor,1.33);
				if(dot(rd,nor)<-0.05)p=rd;//if the ray is incident to surface just reflect
			}
			return normalize(p);
		} 
		return cosPowDir(p,1.0,0.0101-abs(r)*0.01);
	}
}

vec3 color(vec3 ro, vec3 rd) 
{// find color of scene
	vec3 tcol = vec3(0.0),fcol = vec3(1.0);
	minLiteDist=1000.0;
	for( int i=0; i <RayBounces && dot(fcol,fcol)>0.001; i++ )
	{// create light paths iteratively
		vec2 hit = intersect( ro, rd );
        	if( hit.y >= 0.0 ){//hit something
        		ro+= rd * hit.x;// advance ray position
        		vec3 nor = getNormal( ro, hit.y );// get the surface normal
			tcol+=applyLighting(fcol, rd, ro, nor, hit.y);
			rd = getBRDFRay( nor, rd, hit.y );// prepare ray for indirect light gathering (bounce)
		}else{//hit nothing or light so bail
			if(hit.y==-1.0)tcol+=fcol*getBackground( rd );
			else tcol+=fcol*lightColor*2.0;//we hit a light so make it bright!
			break;
		}
	}//light bloom is OR'd in (could be done in post processing)
	return clamp(max(tcol,lightColor/max(lBloom*minLiteDist*minLiteDist,0.5)),0.0,1.0);
}	 



